@@ -20,19 +20,28 @@
 #
 #
 
-from casadi import MX, substitute, Function, vcat, depends_on, vertcat, jacobian, veccat, jtimes, hcat, linspace, DM, constpow, mtimes, vvcat, low, floor, hcat, horzcat, DM, is_equal
+from casadi import MX, substitute, Function, vcat, depends_on, vertcat, jacobian, veccat, jtimes, hcat,\
+                   linspace, DM, constpow, mtimes, vvcat, low, floor, hcat, horzcat, DM, is_equal, \
+                   Sparsity
 from .freetime import FreeTime
 from .direct_method import DirectMethod
 from .multiple_shooting import MultipleShooting
 from .single_shooting import SingleShooting
 from collections import defaultdict
-from .casadi_helpers import DM2numpy, get_meta, merge_meta, HashDict, HashDefaultDict, HashOrderedDict
+from .casadi_helpers import DM2numpy, get_meta, merge_meta, HashDict, HashDefaultDict, HashOrderedDict, HashList, for_all_primitives
 from contextlib import contextmanager
 from collections import OrderedDict
 
 import numpy as np
 from numpy import nan
 
+def transcribed(func):
+    def function_wrapper(self, *args, **kwargs):
+        return func(self._transcribed, *args, **kwargs)
+    function_wrapper._decorator_original = func
+    function_wrapper.__doc__ = func.__doc__
+    return function_wrapper
+
 class Stage:
     """
         A stage is defined on a time domain and has particular system dynamics
@@ -64,16 +73,20 @@ class Stage:
 
         >>> stage = Stage()
         """
-        self.states = []
-        self.qstates = []
-        self.controls = []
-        self.algebraics = []
-        self.parameters = defaultdict(list)
-        self.variables = defaultdict(list)
+        self.states = HashList()
+        self.qstates = HashList()
+        self.controls = HashList()
+        self.algebraics = HashList()
+        self.parameters = defaultdict(HashList)
+        self.variables = defaultdict(HashList)
 
         self._master = parent.master if parent else None
         self.parent = parent
 
+        self._meta = HashDict()
+        self._var_original = None
+        self._var_augmented = None
+
         self._param_vals = HashDict()
         self._state_der = HashDict()
         self._state_next = HashDict()
@@ -83,20 +96,17 @@ class Stage:
         self._initial = HashOrderedDict()
 
         self._placeholders = HashDict()
-        self._placeholder_callbacks = HashDict()
         self._offsets = HashDict()
         self._inf_inert = HashOrderedDict()
         self._inf_der = HashOrderedDict()
-        self._t = MX.sym('t')
+        self._t = self._create_placeholder_expr(0, 't')
         self._stages = []
         self._method = DirectMethod()
         self._t0 = t0
         self._T = T
-        self._public_T = self._create_placeholder_expr(0, 'T')
+        self._public_T  = self._create_placeholder_expr(0, 'T')
         self._public_t0 = self._create_placeholder_expr(0, 't0')
         self._tf = self.T + self.t0
-        if not clone:
-            self._check_free_time()
     
     @property
     def master(self):
@@ -118,23 +128,18 @@ class Stage:
     def tf(self):
         return self._tf
 
-    def _check_free_time(self):
-        if isinstance(self._t0, FreeTime):
-            init = self._t0.T_init
-            self.set_t0(self.variable())
-            self.set_initial(self._t0, init)
-        if isinstance(self._T, FreeTime):
-            init = self._T.T_init
-            self.set_T(self.variable())
-            self.subject_to(self._T>=0)
-            self.set_initial(self._T, init)
-
     def set_t0(self, t0):
         self._t0 = t0
 
     def set_T(self, T):
         self._T = T
 
+    def _param_value(self, p):
+        if p not in self._param_vals:
+            raise Exception("You forgot to declare a value (using ocp.set_value) of the following parameter: " + str(self._meta[p]))
+        else:
+            return self._param_vals[p]
+
     def stage(self, template=None, **kwargs):
         """Create a new :obj:`~rockit.stage.Stage` and add it as to the :obj:`~rockit.ocp.Ocp`.
 
@@ -162,7 +167,7 @@ class Stage:
         self._set_transcribed(False)
         return s
 
-    def state(self, n_rows=1, n_cols=1, quad=False):
+    def state(self, n_rows=1, n_cols=1, quad=False,meta=None):
         """Create a state.
         You must supply a derivative for the state with :obj:`~rockit.stage.Stage.set_der`
 
@@ -193,6 +198,7 @@ class Stage:
         import numpy
         # Create a placeholder symbol with a dummy name (see #25)
         x = MX.sym("x"+str(int(numpy.random.rand()*10000)), n_rows, n_cols)
+        self._meta[x] = merge_meta(meta, get_meta())
         if quad:
             self.qstates.append(x)
         else:
@@ -200,7 +206,7 @@ class Stage:
         self._set_transcribed(False)
         return x
 
-    def algebraic(self, n_rows=1, n_cols=1):
+    def algebraic(self, n_rows=1, n_cols=1,meta=None):
         """Create an algebraic variable
         You must supply an algebraic relation with:obj:`~rockit.stage.Stage.set_alg`
 
@@ -220,28 +226,31 @@ class Stage:
         """
         # Create a placeholder symbol with a dummy name (see #25)
         z = MX.sym("z", n_rows, n_cols)
+        self._meta[z] = merge_meta(meta, get_meta())
         self.algebraics.append(z)
         self._set_transcribed(False)
         return z
 
-    def variable(self, n_rows=1, n_cols=1, grid = ''):
+    def variable(self, n_rows=1, n_cols=1, grid = '',meta=None):
         # Create a placeholder symbol with a dummy name (see #25)
         v = MX.sym("v"+str(np.random.random(1)), n_rows, n_cols)
+        self._meta[v] = merge_meta(meta, get_meta())
         self.variables[grid].append(v)
         self._set_transcribed(False)
         return v
 
-    def parameter(self, n_rows=1, n_cols=1, grid = ''):
+    def parameter(self, n_rows=1, n_cols=1, grid = '',meta=None):
         """
         Create a parameter
         """
         # Create a placeholder symbol with a dummy name (see #25)
         p = MX.sym("p", n_rows, n_cols)
+        self._meta[p] = merge_meta(meta, get_meta())
         self.parameters[grid].append(p)
         self._set_transcribed(False)
         return p
 
-    def control(self, n_rows=1, n_cols=1, order=0):
+    def control(self, n_rows=1, n_cols=1, order=0,meta=None):
         """Create a control signal to optimize for
 
         A control signal is parametrized as a piecewise polynomial.
@@ -277,19 +286,36 @@ class Stage:
             return u
 
         u = MX.sym("u", n_rows, n_cols)
+        self._meta[u] = merge_meta(meta, get_meta())
         self.controls.append(u)
         self._set_transcribed(False)
         return u
 
     def set_value(self, parameter, value):
         if self.master is not None and self.master.is_transcribed:
-            self._method.set_value(self, self.master._method, parameter, value)            
+            def action(parameter, value):
+                self._method.set_value(self, self.master._method, parameter, value)      
         else:
-            self._param_vals[parameter] = value
+            def action(parameter, value):
+                if parameter not in self._meta:
+                    raise Exception("You attempted to set the value of a non-parameter: " + str(parameter))
+                if not np.any([parameter in p for p in self.parameters.values()]):
+                    raise Exception("You attempted to set the value of a non-parameter. Did you mean ocp.set_initial()? Got " + str(parameter))
+                self._param_vals[parameter] = value
+        for_all_primitives(parameter, value, action, "First argument to set_value must be a parameter or a simple concatenation of parameters", rhs_type=DM)
 
-    def set_initial(self, var, value):
+
+    def set_initial(self, var, value, priority=True):
         assert "opti" not in str(var)
-        self._initial[var] = value
+        def action(var, value):
+            if var not in self._meta and var not in self._placeholders:
+                raise Exception("You attempted to set the initial value of an unknown symbol: " + str(var))
+            if np.any([var in p for p in self.parameters.values()]):
+                raise Exception("You attempted to set the initial value of a parameter. Did you mean ocp.set_value()? Got " + str(var))
+            self._initial[var] = value
+            if priority:
+                self._initial.move_to_end(var, last=False)
+        for_all_primitives(var, value, action, "First argument to set_initial must be a variable/signal or a simple concatenation of variables/signals")
         if self.master is not None and self.master.is_transcribed:
             self._method.set_initial(self, self.master._method, self._initial)
 
@@ -300,6 +326,7 @@ class Stage:
         ----------
         state : `~casadi.MX`
             A CasADi symbol created with :obj:`~rockit.stage.Stage.state`.
+            May not be an indexed or sliced state
         der : `~casadi.MX`
             A CasADi symbolic expression of the same size as `state`
 
@@ -314,7 +341,11 @@ class Stage:
         """
         self._set_transcribed(False)
         assert not self._state_next
-        self._state_der[state] = der
+        def action(state, der):
+            if state not in self.states and state not in self.qstates:
+                raise Exception("You used set_der on a non-state: " + str(state))
+            self._state_der[state] = der
+        for_all_primitives(state, der, action, "First argument to set_der must be a state or a simple concatenation of states")
 
     def set_next(self, state, next):
         """Assign an update rule for a discrete state
@@ -337,6 +368,9 @@ class Stage:
         """
         self._set_transcribed(False)
         self._state_next[state] = next
+        def action(state, next):
+            self._state_next[state] = next
+        for_all_primitives(state, next, action, "First argument to set_next must be a state or a simple concatenation of states")
         assert not self._state_der
 
     def add_alg(self, constr):
@@ -364,9 +398,7 @@ class Stage:
                          Note that the final state is not included in this definition
         """
         if grid=='inf':
-            I = self.state(quad=True)
-            self.set_der(I, expr)
-            return self.at_tf(I)
+            return self._create_placeholder_expr(expr, 'integral')
         else:
             return self._create_placeholder_expr(expr, 'integral_control')
 
@@ -556,6 +588,7 @@ class Stage:
         >>> ocp.add_objective( ocp.integral(x) ) # Lagrange term
 
         """
+        assert not self.is_signal(term), "An objective cannot be a signal. You must use ocp.integral or ocp.at_t0/tf to remove the time-dependence"
         self._set_transcribed(False)
         self._objective = self._objective + term
 
@@ -653,31 +686,72 @@ class Stage:
         return depends_on(expr, vertcat(self.x, self.u, self.z, self.t, vcat(self.variables['control']+self.variables['states']), vvcat(self._inf_der.keys())))
 
     def _create_placeholder_expr(self, expr, callback_name):
+        """
+        Placeholders are transcribed in two phases
+           Phase 1: before any decision variables are created
+             e.g. augmenting the state-space (ocp.integral)
+           Phase 2: substituting to concrete decision variables
+
+        """
         r = MX.sym("r_" + callback_name, MX(expr).sparsity())
-        self._placeholders[r] = expr
-        self._placeholder_callbacks[r] = callback_name
+        self._placeholders[r] = (callback_name, expr)
         if self.master is not None:
             self.master._transcribed_placeholders.mark_dirty()
         return r
 
-    def _transcribe_placeholders(self, method, placeholders):
-        for k, v in self._placeholders.items():
-            if k not in placeholders:
-                callback = getattr(method, 'fill_placeholders_' + self._placeholder_callbacks[k])
-                placeholders[k] = callback(self, v)
+    def _transcribe_placeholders(self, phase, method, placeholders):
+        # Fixed-point iteration:
+        # Phase 1 may introduce extra placeholders
+        while True:
+            len_before = len(self._placeholders)
+            for k, (n,v) in list(self._placeholders.items()):
+                if k not in placeholders[phase]:
+                    callback = getattr(method, 'fill_placeholders_' + n)
+                    if phase==2 and k in placeholders[phase-1]:
+                        v = placeholders[phase-1][k]
+                    ret = callback(phase, self, v)
+                    if ret is not None:
+                        placeholders[phase][k] = ret
+                        if phase==2 and isinstance(v,MX) and v.is_symbolic() and k in placeholders[phase-1]:
+                            placeholders[phase][v] = ret
+            len_after = len(self._placeholders)
+            if len_before==len_after: break
 
     # Internal methods
     def _ode(self):
-        ode = veccat(*[self._state_der[k] for k in self.states])
-        quad = veccat(*[self._state_der[k] for k in self.qstates])
+        der = []
+        for k in self.states:
+            try:
+                der.append(self._state_der[k])
+            except:
+                raise Exception("ocp.set_der missing for state defined at " + str(self._meta[k]))
+        ode = veccat(*der)
+        der = []
+        for k in self.qstates:
+            try:
+                der.append(self._state_der[k])
+            except:
+                raise Exception("ocp.set_der missing for quadrature state defined at " + str(self._meta[k]))
+        quad = veccat(*der)
         alg = veccat(*self._alg)
         return Function('ode', [self.x, self.u, self.z, vertcat(self.p, self.v), self.t], [ode, alg, quad], ["x", "u", "z", "p", "t"], ["ode","alg","quad"])
 
     # Internal methods
     def _diffeq(self):
-        next = veccat(*[self._state_next[k] for k in self.states])
-        quad = veccat(*[self._state_next[k] for k in self.qstates])
-
+        val = []
+        for k in self.states:
+            try:
+                val.append(self._state_next[k])
+            except:
+                raise Exception("ocp.set_next missing for state defined at " + str(self._meta[k]))
+        next = veccat(*val)
+        val = []
+        for k in self.qstates:
+            try:
+                val.append(self._state_next[k])
+            except:
+                raise Exception("ocp.set_next missing for quadrature state defined at " + str(self._meta[k]))
+        quad = veccat(*val)
         dt = MX(1,1)
         return Function('ode', [self.x, self.u, vertcat(self.p, self.v), self.t, dt], [next, MX(), quad, MX(0, 1), MX()], ["x0", "u", "p", "t0", "DT"], ["xf","poly_coeff","qf","zf","poly_coeff_z"])
 
@@ -743,9 +817,32 @@ class Stage:
 
     _constr_apply = _expr_apply
 
+    @property
+    def _is_original(self):
+        return not self._var_original
+
+    @property
+    def _original(self):
+        return self._var_original if self._var_original else self
+
+    @property
+    def _augmented(self):
+        return self._var_augmented if self._var_augmented else self
+
     def _set_transcribed(self, val):
         if self.master:
-            self.master._is_transcribed = val
+            if self._is_original:
+                self.master._var_is_transcribed = val
+
+    """
+        In fact, both the original and the augmented should separately kee a transcribed flag
+    """
+    @property
+    def _is_transcribed(self):
+        if self._is_original:
+            return self.master._var_is_transcribed 
+        else:
+            return self._original._is_transcribed
 
     @property
     def is_transcribed(self):
@@ -754,25 +851,26 @@ class Stage:
         else:
             return False
 
-    def _transcribe_recurse(self, pass_nr=1, **kwargs):
+    def _transcribe_recurse(self, phase=1, **kwargs):
         if self._method is not None:
             if self is self.master:
-                self._method.main_transcribe(self, pass_nr=pass_nr, **kwargs)
-            self._method.transcribe(self, pass_nr=pass_nr, **kwargs)
+                self._method.main_transcribe(self, phase=phase, **kwargs)
+            self._method.transcribe(self, phase=phase, **kwargs)
         else:
-            print("master",self)
+            pass
 
         for s in self._stages:
-            s._transcribe_recurse(pass_nr=pass_nr, **kwargs)
+            s._transcribe_recurse(phase=phase, **kwargs)
 
-    def _placeholders_transcribe_recurse(self, placeholders):
+    def _placeholders_transcribe_recurse(self, phase, placeholders):
         if self._method is not None:
-            self._method.transcribe_placeholders(self, placeholders)
+            self._method.transcribe_placeholders(phase, self, placeholders)
 
         for s in self._stages:
-            s._placeholders_transcribe_recurse(placeholders)
+            s._placeholders_transcribe_recurse(phase, placeholders)
 
     def clone(self, parent, **kwargs):
+        assert self._is_original
         ret = Stage(parent, **kwargs)
         from copy import copy, deepcopy
 
@@ -784,10 +882,11 @@ class Stage:
                 subst_to.append(ret.T)
             elif is_equal(k, self.t0):
                 subst_to.append(ret.t0)
+            elif is_equal(k, self.t):
+                subst_to.append(ret.t)
             else:
                 subst_to.append(MX.sym(k.name(), k.sparsity()))
         for k_old, k_new in zip(subst_from, subst_to):
-            ret._placeholder_callbacks[k_new] = self._placeholder_callbacks[k_old]
             ret._placeholders[k_new] = self._placeholders[k_old]
 
         ret.states = copy(self.states)
@@ -818,18 +917,38 @@ class Stage:
             r = r[len(v):]
 
         ret._initial = HashOrderedDict(zip(res[n_constr+1:], self._initial.values()))
-        ret._check_free_time()
 
         if "T" not in kwargs:
             ret._T = copy(self._T)
         if "t0" not in kwargs:
             ret._t0 = copy(self._t0)
-        ret._t = self.t
         ret._method = deepcopy(self._method)
+        ret._method.T = None
+        ret._method.t0 = None
+        ret._var_original = self._var_original
 
-        ret._is_transcribed = False
+        ret._meta = self._meta
+
+        ret._var_is_transcribed = False
         return ret
 
+    def __deepcopy__(self, memo):
+        # Get default deepcopy behaviour
+        import copy
+        deepcopy_method = self.__deepcopy__
+        self.__deepcopy__ = None
+        cp = copy.deepcopy(self, memo)
+        self.__deepcopy__ = deepcopy_method
+        cp.__deepcopy__ = deepcopy_method
+
+        # Custom amendments
+        cp._var_original = self
+        self._var_augmented = cp
+
+        cp._method = self._method
+
+        return cp
+
     def iter_stages(self, include_self=False):
         if include_self:
             yield self
@@ -955,7 +1074,8 @@ class Stage:
             raise Exception(msg)
         N, M = stage._method.N, stage._method.M
 
-        expr_f = Function('expr', [stage.x, stage.z, stage.u], [expr])
+        expr_f = Function('expr', [stage.t, stage.x, stage.z, stage.u], [expr])
+        assert not expr_f.has_free()
 
         time = stage._method.control_grid
         total_time = []
@@ -967,7 +1087,8 @@ class Stage:
             assert tlocal.is_column()
             ts = tlocal[:-1,:]
             for l in range(M):
-                total_time.append(t0+tlocal[:-1])
+                local_t = t0+tlocal[:-1]
+                total_time.append(local_t)
                 coeff = stage._method.poly_coeff[k * M + l]
                 tpower = hcat([constpow(ts,i) for i in range(coeff.shape[1])]).T
                 if stage._method.poly_coeff_z:
@@ -976,7 +1097,7 @@ class Stage:
                     z = mtimes(coeff_z,tpower_z)
                 else:
                     z = nan
-                sub_expr.append(stage._method.eval_at_integrator(stage, expr_f(mtimes(coeff,tpower), z, stage._method.U[k]), k, l))
+                sub_expr.append(stage._method.eval_at_integrator(stage, expr_f(local_t.T, mtimes(coeff,tpower), z, stage._method.U[k]), k, l))
                 t0+=dt
 
         ts = tlocal[-1,:]
@@ -988,11 +1109,11 @@ class Stage:
         else:
             z = nan
 
-        sub_expr.append(stage._method.eval_at_integrator(stage, expr_f(mtimes(stage._method.poly_coeff[-1],tpower), z, stage._method.U[-1]), k, l))
+        sub_expr.append(stage._method.eval_at_integrator(stage, expr_f(time[k+1], mtimes(stage._method.poly_coeff[-1],tpower), z, stage._method.U[-1]), k, l))
 
         return vcat(total_time), hcat(sub_expr)
 
-    def value(self, expr, *args, **kwargs):
+    def value(self, expr):
         """Get the value of an (non-signal) expression.
 
         Parameters
@@ -1059,7 +1180,8 @@ class Stage:
             raise Exception(msg)
         N, M = self._method.N, self._method.M
 
-        expr_f = Function('expr', [self.x, self.z, self.u], exprs)
+        expr_f = Function('expr', [self.t, self.x, self.z, self.u], exprs)
+        assert not expr_f.has_free()
 
         time = vcat(self._method.integrator_grid)
         k = low(self._method.control_grid, t)        
@@ -1087,7 +1209,8 @@ class Stage:
             z = nan
 
         Us = hcat(self._method.U)
-        f = Function(name,[self.gist, t],expr_f.call([mtimes(coeff,tpower), z, Us[:,k]]), options)
+        f = Function(name,[self.gist, t],expr_f.call([t, mtimes(coeff,tpower), z, Us[:,k]]), options)
+        assert not f.has_free()
 
         if numpy:
             def wrapper(gist, t):
